# @(#) $Id$
# $Source$
#
# $Copyright$

## @class
# @brief This is a Windows specific instance of the Remote Host Explorer
#
# This class understands how to remotely discovery Windows hosts
package VRTS::HAL::RH::RHExplorer::Windows;

use strict;
use warnings;
use Carp;
use VRTS::HAL::mkpath;
use VRTS::HAL::Logger qw(UMI);

use VRTS::HAL::RH::RHExplorer qw(/^K_/ /^V_/ /^S_/);
use vars qw(@ISA);
@ISA = qw( VRTS::HAL::RH::RHExplorer );

## @method void validate_control_defaults(hashref)
# Validate initial control section parameters with Windows
# specific values.
# Called only in an initialization context and prior to the invocation
# of the initialize_environment() method.
# @param hashref Reference to hash with initialization arguments.
#   Update the hash with defaults.
# @exception dies On error
sub validate_control_defaults
{
    my $self = shift;
    my $hr = shift;
    if ( !exists($hr->{K_NETWORK_PROTOCOL.""}) ) {
	$hr->{K_NETWORK_PROTOCOL.""} = V_NP_WMI."";
    }
    if ( $hr->{K_COLLECTION_MODE.""} eq V_CM_REMOTE."" ) {
	die "Cannot configure Windows remote discovery on non-Windows platforms\n"
	    if $^O !~ m/Win/;
    }
    my $np = $hr->{K_NETWORK_PROTOCOL.""};
    die "Cannot access Windows hosts using SSH\n" if $np eq V_NP_SSH."";
}

## @method void initialize_environment(args)
# Method to customize the explorer at definition time.
# @param args Arguments provided to the constructor.
sub initialize_environment
{
    my $self = shift;
    my $dbg = $self->{Debug};
    my $ostype = $self->{OSType};
    $self->{RHE_log}->debug1("RHE_W::$ostype initialize_environment()");
}


## @method void do_remote_discovery($oc,$secs)
# Method to discover the data corresponding to a list of sections.
# @param oc A VRTS::HAL::RH::ObjectCache object to populate.
# @param secs Reference to a list of section names.
# @exception dies on error
sub do_remote_discovery
{
    my $self = shift;
    my $oc = shift;
    my $secs = shift;

    my $dbg = $self->{Debug};

    # Uncomment the following line to write all of our debug logging to STDOUT
    # without having to run with -x, making everything else verbose.
    # 'debug2' == 7, and it writes 7 (debug2)-->1 (critical) to STDOUT
    # $self->{RHE_log}->set_print_level(*STDOUT, 'debug2');

    my $csec = $self->control_section();
    my $account = $csec->{K_ACCOUNT_NAME.""};
    my $passwd = $csec->{K_ACCOUNT_PASSWORD.""}; # encrypted
    my $host = $csec->{K_HOST_NAME.""};
    my $coll = $self->parent();
    my $hk = $coll->{HostId};
    my ($vol, $d, $file) = File::Spec->splitpath($self->{ConfigFile});
    my $sec = VRTS::HAL::RH::RHExplorer::S_CONTROL;
    my $val = K_ACCOUNT_PASSWORD;
    my $mhobf = join(':', $hk, $file, $sec, $val, $passwd);
    local($ENV{HALMHOBF}) = $mhobf;
    my $dnu = $csec->{"DoNotDeleteDataFile"};

    # we need to impersonate
    my $coll_csec = $coll->control_section();
    my $imp_account = $coll_csec->{VRTS::HAL::RH::RHCollector::K_IMP_ACCOUNT.""};
    $val = VRTS::HAL::RH::RHCollector::K_IMP_ACCOUNT_PASS;
    my $imp_passwd = $coll_csec->{$val};
    die "Local impersonation account not defined in collector.\n"
	if ( !defined($imp_account) || !defined($imp_passwd) );
    ($vol, $d, $file) = File::Spec->splitpath($coll->{ConfigFile});
    my $imp_mhobf = join(':', $hk, $file, $sec, $val, $imp_passwd);
    local($ENV{IMP_HALMHOBF}) = $imp_mhobf;
    local($ENV{IMP_USER}) = $imp_account;

    $self->{RHE_log}->debug("RHE_W::$self->{OSType} Process ID -- $$..");
    $self->{RHE_log}->debug("RHE_W::$self->{OSType} account -- $account");
    $self->{RHE_log}->debug("RHE_W::$self->{OSType} password -- $passwd");
    $self->{RHE_log}->debug("RHE_W::$self->{OSType} host -- $host");
    $self->{RHE_log}->debug("RHE_W::$self->{OSType} mhobf -- $mhobf");
    $self->{RHE_log}->debug("RHE_W::$self->{OSType} imp_account -- $imp_account");
    $self->{RHE_log}->debug("RHE_W::$self->{OSType} imp_mhobf -- $imp_mhobf");

    # directory where vbscript output data file will be placed.
    my $dd = $self->get_datadir();
    $self->{RHE_log}->debug("RHE_W::$self->{OSType} data directory -- '$dd'");

    # find halsecwrapper.exe, the utility to decrypt the password.
    my $halsec = &hal_mkpath_exe(HAL_PT_CC_INSTALL, "bin/halsecwrapper.exe");

    # find the data-gathering vbscript that runs commands via WMI 
    my $script = &hal_mkpath(HAL_PT_CC_INSTALL, "lib/RHWindowsScript.vbs");

    # pass in script and remote directory name.
    # To make remote directory name unique, we use user and collector host name.
    #my $rdir = VRTS::HAL::RH::RHCollector::fs_safe_hostname($coll->{Hostname});
    # $rdir = $rdir . "_" . $account;
    # $self->{RHE_log}->debug("RHE_W::$self->{OSType} rdir -- $rdir");

    # Generate output file name based on hostname and UTC time
    my $utc = time();
    my ($second,$min,$hour,$mday,$mon,$year)=(gmtime($utc))[0..5];
    my $ts = sprintf("%4u%02u%02u%02u%02u", $year+1900,$mon+1,$mday,$hour,$min);
    my $ofile = $host . "_Win_" . $ts . ".txt";
    $self->{RHE_log}->debug("RHE_W::$self->{OSType} ofile -- '$ofile'");

    my $oPathnFile = "$dd/$ofile";
    my $oPathnFile2 = "$dd/utf8_$ofile";
    $self->{RHE_log}->debug("RHE_W::$self->{OSType} vbscript datafile '$oPathnFile'");

    my $cmd;
    # construct the command, including the halsecwrapper password unencryption
    # The //NOLOGO and //B flags modify the way cscript acts.  The //NOLOGO
    # flag tells cscript to not display its header.  The //B flag tells
    # cscript to run the command in batch mode, that is, with stdout and
    # stderr redirected to NULL.
    $cmd = "\"$halsec\" cscript //NOLOGO //B \"$script\"";
    $cmd .= " /u $account /p REPLACEPASS /s $host /d \"$dd\"";
    $cmd .= " /f \"$ofile\"";
    # if dbg, tell vbs script to dump additional info into datafile.
    $cmd .= " /v" if $dbg;
    $self->{RHE_log}->debug("RHE_W::cmd: '$cmd'");
    my $rc = system($cmd);
    
    # The discovery script returns 1 when the discovery utterly fails.
    # Inability to connect to the WMI namespaces ensures that no discovery
    # can occur, so it is a fatal error.  Other non-zero return codes 
    # indicate partial discovery but aren't fatal.  Abort if we get 1.
    # The byte above the low byte contains the return code, so roll right
    # 8 places.
    if (($rc >> 8) == 1) { 
        $self->{RHE_log}->error(&UMI(86,18,71), 
               "Could not connect to $host Windows Management Interface (WMI) Namespace(s)");
        die "Could not connect to Windows Management Interface (WMI) Namespace(s)";
    }
    # convert the utf16 unicode file to utf8
    open (IFILE, "<:raw:encoding(UTF-16LE)", $oPathnFile) or die "could not read  $oPathnFile";
    open (OFILE, ">:raw:encoding(UTF-8)", $oPathnFile2) or die "could not write  $oPathnFile2";
    while (<IFILE>){
    print OFILE  $_;
    }
    close OFILE;
    close IFILE;
    

    my $fp = VRTS::HAL::RH::ScriptParser->new(Filename=>$oPathnFile2,
	                                      Debug =>$dbg);
    $fp->{host} = $host;
    eval { $fp->parse($oc) };
    my $exception = $@;
    
    # When parsing is finished, remove the output data file unless the 
    # "DoNotDeleteDataFile" attribute in the Control section of the host's
    # config file is set to 1.
    {
	no warnings;
        $self->{RHE_log}->debug("RHE_W::$self->{OSType} DNU -- $dnu");
	undef $fp; # close file handle
        unlink $oPathnFile unless (defined($dnu) && $dnu == 1);
        unlink $oPathnFile2 unless (defined($dnu) && $dnu == 1);
    }
    die $exception if $exception;
    return;
}

## @method void do_probe($sec, $cfg) 
# @param $sec The section of the config file to be parsed.
# @param $cfg The configuration file that is to be parsed to retrieve values 
# for probing.
# do_probe() invokes RHWindowsScript.vbs with a special command-line flag
# that directs it to perform a probe of the specified host.  The script runs
# a set of tests to determine whether the specified account and credentials
# will be sufficient to fully discover the host.  The tests currently consist
# of attempted connection to the two required WMI namespaces (CIMV2 and WMI),
# fetching two classes, one requiring admin privileges (Win32_DiskDrive), one
# normal (Win32_OperatingSystem), and mapping of the admin$ share.  In the
# process of fetching the OS class, the script obtains the OS Version (2000,
# 2003, 2008, etc.)
#
# Passing tests are marked "OK".  Failing tests are marked "NOK", and any
# relevant error number appended.
# If all tests pass, the deviceState attribute is set to "NORMAL".  If any
# fail, it is set to "ERROR".
sub do_probe { 
    my $self = shift; 
    my $section = shift; 
    my $cfg= shift;

    my $dbg = $self->{Debug};

    # Uncomment the following line to write all of our debug logging to STDOUT
    # without having to run with -x, making everything else verbose.
    # 'debug2' == 7, and it writes 7 (debug2)-->1 (critical) to STDOUT
    # $self->{RHE_log}->set_print_level(*STDOUT, 'debug2');

    $self->{RHE_log}->debug1("RHE_W::In Windows do_probe()");
    $self->{RHE_log}->debug1("RHE_W::  section: '$section'");

    my $account = $cfg->get_string_value(S_CONTROL, K_ACCOUNT_NAME, undef);
    my $passwd = $cfg->get_string_value(S_CONTROL, K_ACCOUNT_PASSWORD, undef); # encrypted
    my $host = $cfg->get_string_value(S_CONTROL, K_HOST_NAME, undef);
    my $coll = $self->parent();
    my $hk = $coll->{HostId};
    my ($vol, $d, $file) = File::Spec->splitpath($self->{ConfigFile});
    my $sec = VRTS::HAL::RH::RHExplorer::S_CONTROL;
    my $val = K_ACCOUNT_PASSWORD;
    my $mhobf = join(':', $hk, $file, $sec, $val, $passwd);
    local($ENV{HALMHOBF}) = $mhobf;
    my $dnu = $cfg->get_string_value(S_CONTROL, "DoNotDeleteDataFile", undef);

    # we need to impersonate
    my $coll_csec = $coll->control_section();
    my $imp_account = $coll_csec->{VRTS::HAL::RH::RHCollector::K_IMP_ACCOUNT.""};
    $val = VRTS::HAL::RH::RHCollector::K_IMP_ACCOUNT_PASS;
    my $imp_passwd = $coll_csec->{$val};
    die "Local impersonation account not defined in collector.\n"
	if ( !defined($imp_account) || !defined($imp_passwd) );
    ($vol, $d, $file) = File::Spec->splitpath($coll->{ConfigFile});
    my $imp_mhobf = join(':', $hk, $file, $sec, $val, $imp_passwd);
    local($ENV{IMP_HALMHOBF}) = $imp_mhobf;
    local($ENV{IMP_USER}) = $imp_account;

    # directory where vbscript output data file will be placed.
    my $dd = $self->get_datadir();
    $self->{RHE_log}->debug("RHE_W::$self->{OSType} data directory -- '$dd'");

    # find halsecwrapper.exe, the utility to decrypt the password.
    my $halsec = &hal_mkpath_exe(HAL_PT_CC_INSTALL, "bin/halsecwrapper.exe");

    # find the data-gathering vbscript that runs commands via WMI 
    my $script = &hal_mkpath(HAL_PT_CC_INSTALL, "lib/RHWindowsScript.vbs");

    if ($dbg) {
        my ($key, $keyval);
        my $keys = $cfg->enumerate_keys($section); 
        foreach $key (@{$keys}) { 
            $keyval = $cfg->get_string_value($section, $key, undef); 
            $self->{RHE_log}->debug("    key '$key', value '$keyval'");
        }
    }

    my $utc = time();
    my ($second,$min,$hour,$mday,$mon,$year)=(gmtime($utc))[0..5];
    my $ts = sprintf("%4u%02u%02u%02u%02u", $year+1900,$mon+1,$mday,$hour,$min);
    my $ofile = $host . "_Win_" . $ts . "_probe.txt";
    $self->{RHE_log}->debug1("RHE_W::$self->{OSType} ofile -- '$ofile'");

    my $oPathnFile = "$dd/$ofile";
    my $oPathnFile2 = "$dd/utf8_$ofile";
    $self->{RHE_log}->debug("RHE_W::$self->{OSType} vbscript datafile '$oPathnFile'");

    # return the probe results in an anonymous hash.
    my $result = {};
    $result->{OSType} = "Windows";
    $result->{deviceState} = "NORMAL";  # until we find out otherwise...

    my $cmd;
    # See the comment in do_remote_discovery for command line explanation.
    # Also specify /i (probe) flag to tell vscript to perform a probe.
    $cmd = "\"$halsec\" cscript //NOLOGO //B \"$script\"";
    $cmd .= " /u $account /p REPLACEPASS /s $host /d \"$dd\"";
    $cmd .= " /f \"$ofile\" /i";
    # if dbg, tell vbs script to dump additional info into datafile.
    $cmd .= " /v" if $dbg;
    $self->{RHE_log}->debug("RHE_W::cmd: '$cmd'");
    my $sysres = system($cmd);
    $self->{RHE_log}->debug("RHE_W::system($cmd) result '$sysres'");
    if ($sysres != 0) {
        my $perr = sprintf("%x", $sysres);
        $self->{RHE_log}->debug("RHE_W::We're outa here ($perr)!!");
        $result->{deviceState} = "ERROR";
        $result->{msgs} = "ERROR: probe result $sysres (0x$perr)";
        return($result);
    }

    # convert the utf16 unicode file to utf8
    open (IFILE, "<:raw:encoding(UTF-16LE)", $oPathnFile) or die "could not read  $oPathnFile";
    open (OFILE, ">:raw:encoding(UTF-8)", $oPathnFile2) or die "could not write  $oPathnFile2";
    while (<IFILE>){
    print OFILE  $_;
    }
    close OFILE;
    close IFILE;

    if (open FILE, "<", $oPathnFile2) {
        $self->{RHE_log}->debug("RHE_W::open probefile '$oPathnFile' successful");
        $result->{msgs} = "";
        $result->{deviceState} = "";
        my @lines = <FILE>;
        for (my $i = 0; $i < @lines; $i++) {
            chomp($lines[$i]);
            $self->{RHE_log}->debug("RHE_W::Line($i): '$lines[$i]'");
	    # line format is: "<attr>=<value>", so split on "="
	    my ($attr,$value) = split(/=/, $lines[$i]);
            $self->{RHE_log}->debug1("  attr: '$attr', value: '$value'");
            $result->{$attr} = $value;
        }
	$self->_diagnose_probe($result);
        close(FILE);
        unlink $oPathnFile unless (defined($dnu) && $dnu == 1);
        unlink $oPathnFile2 unless (defined($dnu) && $dnu == 1);
    } else {
        $self->{RHE_log}->debug("RHE_W::open probefile '$oPathnFile2' FAILED");
    }

    return($result);
} 


## @method private void _diagnose_probe($r_result)
# _diagnose_probe examines the results of the probe operation, and diagnoses
# their ramifications for the successful exploration of the host, setting 
# 'deviceState' and 'msgs' appropriately.
#
# @param $r_result reference to result hash
#
sub _diagnose_probe
{
    my $self = shift;
    my $r_result = shift;

    # First detect and handle the fatal errors.

    # If we can't connect to the WMI Namespaces, we can't get any data at all.
    # Check this first: it is THE most fatal error.
    if (($r_result->{CIMV2_Namespace} =~ "NOK") ||  
        ($r_result->{WMI_Namespace} =~ "NOK")) {
        $self->{RHE_log}->error(&UMI(86,18,71), 
                    "Could not connect to Windows Management Interface (WMI) Namespace(s)");
        $r_result->{msgs} = "Could not connect to Windows Management Interface (WMI) Namespace(s): Please check that type of host operating system is correct (Windows was specified), that the hostname is valid and reachable, the credentials are valid for remote host's local Administrator account, and ensure that WMI service is running on remote host";
        $r_result->{deviceState} = "ERROR";
        return;
    }

    # Next, check whether this OS is supported.  Reject Windows 2000.
    if ($r_result->{OSVersion} eq "2000") {
        $self->{RHE_log}->error(&UMI(86,18,34), 
                    "Unsupported Operating System version: Windows $r_result->{OSVersion}");
        $r_result->{msgs} = 
                    "Unsupported Operating System version (Windows $r_result->{OSVersion})";
        $r_result->{deviceState} = "ERROR";
        return;
    }

    # Another fatal error is the inability to fetch the DiskDrive Class, even
    # though we can connect to the namespaces, and perhaps even fetch other
    # (less privileged) classes.  Although we *could* attempt discovery, the
    # results would be singularly unsatisfactory, so let's not try.
    if ($r_result->{DiskDriveClass} eq "NOK") {
        $r_result->{msgs} = 
               "Insufficient Windows Management Instrumentation (WMI) permissions: cannot fetch DiskDrive Class"; 
        $r_result->{deviceState} = "ERROR";
        $self->{RHE_log}->error(&UMI(86,18,73), 
               "Insufficient Windows Management Instrumentation (WMI) permissions: cannot fetch DiskDriveClass"); 
        return;
    }

    # Now detect and handle the warnings.

    # If not Windows 2008, we require fcinfo in order to perform FC discovery.
    # Therefore, warn if fcinfo isn't found, unless the host is a VMWare 
    # Virtual Machine, which don't have FC data (no HBAs or LUNs).
    if (($r_result->{OSVersion} ne "2008") && ($r_result->{fcinfo} ne "OK")) { 
        if ($r_result->{IsVirtualMachine} ne "True") {
            $r_result->{deviceState} = "WARNING";
            $r_result->{msgs} = 
               "fcinfo.msi not installed: no Fibre Channel discovery possible";
            $self->{RHE_log}->error(&UMI(86,18,72), 
               "fcinfo.msi not installed: no Fibre Channel discovery possible");
        }
    }

    # Now check whether any attrs are NOK.  If so, we'll set a WARNING.
    my $someNOK = 0;
    foreach my $attr (sort keys(%$r_result)) {
        if ($r_result->{$attr} =~ "NOK") {
           $someNOK = 1;
        }
    }

    # We we haven't yet set deviceState, look at the results of the tests.
    # If any were NOK, set deviceState to "WARNING"; else to "NORMAL".
    if ($r_result->{deviceState} eq "") {
        if ($someNOK == 0) {
            $r_result->{deviceState} = "NORMAL";
            if ($r_result->{msgs} eq "") {  # don't overwrite if set
                $r_result->{msgs} = "Probe Successful";
            }
        } else {
            $r_result->{deviceState} = "WARNING";
            $r_result->{msgs} = "Probe Succeeded Conditionally";
        }
    }

    return;
}


1;
